package core.services;

import core.api.models.account.AccountDeleteRequest;
import core.api.models.account.EmailChangeRequest;
import core.api.models.account.PasswordChangeRequest;
import core.api.models.users.NtfyInformation;
import core.api.models.users.UserAddInfo;
import core.api.models.users.UserInfo;
import core.api.models.users.UserUpdateInfo;
import core.entities.RoleEntity;
import core.entities.User;
import core.entities.UserRole;
import core.entities.timemanager.Taskgroup;
import core.repositories.RoleRepository;
import core.repositories.UserRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;

import javax.transaction.Transactional;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Optional;

@Service
public class UserService {

    @Autowired
    private UserRepository userRepository;

    @Autowired
    private RoleRepository roleRepository;
    @Autowired private TaskgroupService taskgroupService;

    public List<UserInfo> getAllUserInfos() {
        Iterable<User> users = userRepository.findAll();
        List<UserInfo> userInfos = new ArrayList<>();
        for(User user : users) {
            userInfos.add(new UserInfo(user));
        }
        return userInfos;
    }

    @Transactional
    public boolean deleteUserByName(String username) {
        if(userRepository.existsByUsername(username)) {
            userRepository.deleteByUsername(username);
            return true;
        }
        return false;
    }

    public boolean updateUser(String username, UserUpdateInfo userUpdateInfo) {
        Optional<User> user_req = userRepository.findByUsername(username);
        if(user_req.isPresent()) {
            User user = user_req.get();
            if(!userUpdateInfo.getPassword().isBlank()) {
                String hashedPassword = new BCryptPasswordEncoder().encode(userUpdateInfo.getPassword());
                user.setPassword(hashedPassword);
            }
            user.setEmail(userUpdateInfo.getEmail());

            Optional<RoleEntity> adminRoleEntity = roleRepository.findByName(UserRole.ROLE_ADMIN);
            if(userUpdateInfo.isAdmin()) {
                adminRoleEntity.ifPresent(adminRole -> {
                    user.getRoles().add(adminRole);
                });
            } else {
                adminRoleEntity.ifPresent(adminRole -> {
                    user.getRoles().remove(adminRole);
                });
            }

            userRepository.save(user);
            return true;
        }
        return false;
    }

    public boolean addUser(UserAddInfo userAddInfo) {
        if(userRepository.existsByUsername(userAddInfo.getUsername()) || userRepository.existsByEmail(userAddInfo.getEmail())) {
            return false;
        }

        String hashedPassword = new BCryptPasswordEncoder().encode(userAddInfo.getPassword());
        User user = new User(userAddInfo.getUsername(), userAddInfo.getEmail(), hashedPassword);
        Optional<RoleEntity> adminRoleEntity = roleRepository.findByName(UserRole.ROLE_ADMIN);
        Optional<RoleEntity> userRoleEntity = roleRepository.findByName(UserRole.ROLE_USER);
        userRoleEntity.ifPresent(userRole -> user.getRoles().add(userRole));
        if(userAddInfo.isAdmin()) {
            adminRoleEntity.ifPresent(adminRole -> user.getRoles().add(adminRole));
        }
        userRepository.save(user);
        return true;
    }

    public Optional<UserInfo> getUserByName(String username) {
        Optional<User> requestedUser = userRepository.findByUsername(username);
        return requestedUser.map(UserInfo::new);
    }

    public int updateEmailOfUsername(String username, EmailChangeRequest emailChangeRequest) {
        Optional<User> requestedUser = userRepository.findByUsername(username);
        if(requestedUser.isEmpty()) {
            return 1;
        }

        if(!new BCryptPasswordEncoder().matches(emailChangeRequest.getPassword(), requestedUser.get().getPassword())) {
            return 2;
        }

        User user = requestedUser.get();
        user.setEmail(emailChangeRequest.getEmail());
        userRepository.save(user);

        return 0;
    }

    public int updatePassword(String username, PasswordChangeRequest passwordChangeRequest) {
        Optional<User> requestedUser = userRepository.findByUsername(username);
        if(requestedUser.isEmpty()) {
            return 1;
        }

        BCryptPasswordEncoder passwordEncoder  = new BCryptPasswordEncoder();

        if(!passwordEncoder.matches(passwordChangeRequest.getOldPassword(), requestedUser.get().getPassword())) {
            return 2;
        }

        User user = requestedUser.get();
        user.setPassword(passwordEncoder.encode(passwordChangeRequest.getNewPassword()));
        userRepository.save(user);

        return 0;
    }

    @Transactional
    public int deleteUserVerfied(String username, AccountDeleteRequest accountDeleteRequest) {
        Optional<User> requestedUser = userRepository.findByUsername(username);
        if(requestedUser.isEmpty()) {
            return 1;
        }

        BCryptPasswordEncoder passwordEncoder  = new BCryptPasswordEncoder();

        if(!passwordEncoder.matches(accountDeleteRequest.getPassword(), requestedUser.get().getPassword())) {
            return 2;
        }

        taskgroupService.deleteTaskgroupByUser(requestedUser.get());
        userRepository.deleteByUsername(username);
        return 0;
    }

    public void editNtfyInformation(User user, NtfyInformation ntfyInformation) {
        user.setNtfy_host(ntfyInformation.getNtfy_host());
        user.setNtfy_topic(ntfyInformation.getNtfy_topic());
        user.setNtfy_username(ntfyInformation.getNtfy_user());
        user.setNtfy_token(ntfyInformation.getNtfy_token());

        userRepository.save(user);
    }
}
